/******************************************************************************
 *
 * Copyright (c) 2019, the nbcelltest authors.
 *
 * This file is part of the nbcelltest library, distributed under the terms of
 * the Apache License 2.0.  The full license can be found in the LICENSE file.
 *
 */
div.CelltestTool {

}

div.CelltestsControls {
    display: flex;
    flex-direction: column;
    padding: 7px;
}

div.CelltestsControls > label {
    font-size: 13px;
    text-transform: uppercase;
    width: 100%;
    margin-bottom: 5px;
}

div.CelltestsControls > div {
    display: flex;
    flex-direction: row;
    width: 100%;
}

div.CelltestsControls > div > button {
    width: 50%;
}
div.CelltestsControls > div > button:hover {
    background: var(--jp-layout-color2);
}

div.CelltestsRules {
    display: flex;
    flex-direction: column;
    padding: 7px;
}

div.CelltestsRules > label {
    font-size: 13px;
    text-transform: uppercase;
    width: 100%;
    margin-bottom: 5px;
}


div.CelltestsRules > div {
    display: flex;
    flex-direction: column;
}

div.CelltestsRules > div > div {
    display: flex;
    flex-direction: row;
}

div.CelltestsRules > div > div > span,
div.CelltestsRules > div > div > input[type="number"] {
    flex: 1;
    width: 45%;
}

div.CelltestsEditor {
    display: flex;
    flex-direction: column;
    background: cyan;
    margin-bottom: 5px;
}


div.CelltestsEditor div.CodeMirror-lines {
    border: 1px solid lightgrey;
}